import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)

import tensorflow as tf
import tensorflow_probability as tfp
import keras.backend as K
tfd = tfp.distributions


class Losses:
    def __init__(self,
                 loss_fn,
                 observation_noise,
                 beta=0,
                 z_mean=None,
                 z_sigma=None,
                 z2_mean=None,
                 z2_log_sigma=None,
                 z_mean_TD=None,
                 z_log_sigma_TD=None):
        self.loss = self._get_loss(loss_fn)
        self.beta = beta
        self.observation_noise = observation_noise
        self.z2_mean = z2_mean
        self.z2_log_sigma = z2_log_sigma
        self.z_mean = z_mean
        self.z_sigma = z_sigma
        self.z_mean_TD = z_mean_TD
        self.z_log_sigma_TD = z_log_sigma_TD
        if self.z_mean != None and self.z_sigma != None:
            assert z_mean_TD != None, "z_mean_TD should be defined!"
            assert z_log_sigma_TD != None, "z_log_sigma_TD should be defined!"

    def _get_loss(self, loss_fn):
        losses = {
            "binary": self._binary,
            "normal": self._normal,
            "normalDiag": self._normalDiag,
            "bernoulli": self._bernoulli
        }
        return losses[loss_fn]

    """
      Making it custom metric to be able to feed it to Keras API - actually no need for y_true, y_pred
    """
    def KL_divergence(self, y_true, y_pred):
        if self.z2_mean == None:
            return 0.0
        p2 = tfd.Normal(self.z2_mean, tf.exp(self.z2_log_sigma) + 1e-12)
        q2 = tfd.Normal(0, 1)
        if self.z_mean != None and self.z_sigma != None:
            p = tfd.Normal(self.z_mean, self.z_sigma + 1e-12)
            q = tfd.Normal(self.z_mean_TD, tf.exp(self.z_log_sigma_TD) + 1e-12)
            return self.beta * (tf.reduce_mean(tfd.kl_divergence(p, q)) +
                                tf.reduce_mean(tfd.kl_divergence(p2, q2)))
        else:
            return self.beta * tf.reduce_mean(tfd.kl_divergence(p2, q2))

    """
      For binarized input with KL term (!)
    """
    def _binary(self, x_true, x_reco):
        return -tf.nn.sigmoid_cross_entropy_with_logits(
            labels=x_true, logits=x_reco) + self.KL_divergence(None, None)

    def _bernoulli(self, x_true, x_reco):
        return -tf.reduce_mean(tfd.Bernoulli(
            logits=x_reco)._log_prob(x_true)) + self.KL_divergence(None, None)

    """
      For non binarized input with KL term(!)
    """
    def _normal(self, x_true, x_reco):
        return -tf.reduce_mean(
            tfd.Normal(x_reco, scale=self.observation_noise)._log_prob(
                x_true)) + self.KL_divergence(None, None)

    def _normalDiag(self, x_true, x_reco):
        return -tf.reduce_mean(
            tfd.MultivariateNormalDiag(
                x_reco, scale_identity_multiplier=self.observation_noise).
            _log_prob(x_true)) + self.KL_divergence(None, None)



from keras.preprocessing.image import ImageDataGenerator
import numpy as np
from .data_loader import DataLoader
import os


class DataGeneratorWithLabels:
    def __init__(self,
                 image_shape,
                 batch_size,
                 file_path=os.getcwd() + '/csnl/data/scramtex_700_28px.pkl',
                 binarize=False,
                 whiten=False,
                 contrast_normalize=False):
        self.DATA_LOADER = DataLoader(file_path, binarize, whiten,
                                      contrast_normalize)
        self.IMAGE_SHAPE = image_shape
        self.BATCH_SIZE = batch_size
        self.DATA_GENERATOR = ImageDataGenerator()
        self.TRAIN, self.TRAIN_LABELS, self.TEST, self.TEST_LABELS = self.DATA_LOADER.get_data_and_labels(
        )
        print("Train SHAPE : ", self.TRAIN.shape)
        means = {
            0: [],
            1: [],
            2: [],
            3: [],
            4: [],
            5: [],
            6: [],
            7: [],
            8: [],
            9: []
        }
        stds = {
            0: [],
            1: [],
            2: [],
            3: [],
            4: [],
            5: [],
            6: [],
            7: [],
            8: [],
            9: []
        }
        for ind, label in enumerate(self.TRAIN_LABELS):
            means[label].append(self.TRAIN[ind])
        for K in means.keys():
            stds[K] = np.std(means[K])
            means[K] = np.mean(means[K])
        print("MEAN : ", means)
        print("STD : ", stds)

        print("Test SHAPE : ", self.TEST.shape)
        means = {
            0: [],
            1: [],
            2: [],
            3: [],
            4: [],
            5: [],
            6: [],
            7: [],
            8: [],
            9: []
        }
        stds = {
            0: [],
            1: [],
            2: [],
            3: [],
            4: [],
            5: [],
            6: [],
            7: [],
            8: [],
            9: []
        }
        for ind, label in enumerate(self.TEST_LABELS):
            means[label].append(self.TEST[ind])
        for K in means.keys():
            stds[K] = np.std(means[K])
            means[K] = np.mean(means[K])
        print("MEAN : ", means)
        print("STD : ", stds)

    def flow(self):
        return self.DATA_GENERATOR.flow(self.TRAIN,
                                        self.TRAIN_LABELS,
                                        batch_size=self.BATCH_SIZE)



import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)

from ..losses import Losses
from abc import abstractmethod
from keras.layers import Input, Dense, ReLU, Lambda, Add
from keras.models import Model
import keras.backend as K
from keras.optimizers import RMSprop, Adam
import tensorflow as tf
import tensorflow_probability as tfp
tfd = tfp.distributions


class LadderVAE:
    def __init__(self,
                 input_shape,
                 latent_dim1,
                 latent_dim2,
                 mean_sigma_input_shape=256):
        self.input_shape = input_shape
        self.latent_dim1 = latent_dim1
        self.latent_dim2 = latent_dim2
        self.BATCH_SIZE = self.input_shape[0]
        self._mean_sigma_input_shape = mean_sigma_input_shape  # arbitrary

    @abstractmethod
    def encoder1(self):
        pass

    @abstractmethod
    def encoder2(self):
        pass

    @abstractmethod
    def decoder2(self):
        pass

    @abstractmethod
    def decoder1(self):
        pass

    def mean_log_sigma_model(self):
        inp = Input(shape=(self._mean_sigma_input_shape, ))
        mean, log_sigma = Dense(self.latent_dim1)(inp), Dense(
            self.latent_dim1)(inp)
        model = Model(inp, [mean, log_sigma], name="mean_log_sigma_model")
        return model

    def _get_sigma(self, args):
        log_sigma1, log_sigma2 = args
        return K.pow(
            K.pow(K.exp(log_sigma1) + 1e-12, -2) +
            K.pow(K.exp(log_sigma2) + 1e-12, -2) + 1e-12, -1)

    def _get_mean(self, args):
        mean1, log_sigma1, mean2, log_sigma2 = args
        return (mean1 * K.pow(K.exp(log_sigma1) + 1e-12, -2) + mean2 *
                K.pow(K.exp(log_sigma2) + 1e-12, -2)) * self._get_sigma(
                    [log_sigma1, log_sigma2])

    def _get_sigma_gen(self, args):
        log_sigma1 = args
        return K.pow(K.exp(log_sigma1), 2)

    def _get_mean_gen(self, args):
        mean1, log_sigma1 = args
        return mean1 * K.pow(K.exp(log_sigma1) + 1e-12,
                             -2) * self._get_sigma_gen(log_sigma1)

    def _sample(self, args):
        z_mean, z_sigma = args
        dist = tfd.Normal(loc=z_mean, scale=z_sigma)
        return dist.sample()

    def _reparametrize(self, args):
        z_mean, z_log_sigma = args
        epsilon = K.random_normal(shape=(self.BATCH_SIZE, self.latent_dim2),
                                  mean=0,
                                  stddev=1)
        return z_mean + K.exp(z_log_sigma) * epsilon

    def get_compiled_model(self, *args):
        loss_fn, lr, decay, self.observation_noise, self.beta = args
        input_img = Input(batch_shape=self.input_shape)

        encoder1 = self.encoder1()
        encoder2 = self.encoder2()
        decoder1 = self.decoder1()
        decoder2 = self.decoder2()
        mean_log_sigma_model_for_top_down_calc = self.mean_log_sigma_model()

        d1 = encoder1(input_img)
        d2 = encoder2(d1)

        # Reparametrization deepest layer
        self.z2_mean, self.z2_log_sigma = Dense(self.latent_dim2,
                                                name="mean_z2")(d2), Dense(
                                                    self.latent_dim2,
                                                    name="log_sigma_z2")(d2)

        self.z2 = Lambda(self._reparametrize,
                         name="sampling_z2")([self.z2_mean, self.z2_log_sigma])

        # Top down and bottom up mean and stddev calculation
        self.z1_intermediate = decoder2(self.z2)

        self.z1_mean_TD, self.z1_log_sigma_TD = mean_log_sigma_model_for_top_down_calc(
            self.z1_intermediate)

        self.z1_mean_BU, self.z1_log_sigma_BU = Dense(
            self.latent_dim1,
            name="bottom_up_mean")(d1), Dense(self.latent_dim1,
                                              name="bottom_up_log_sigma")(d1)

        # Combine mean and sigma
        self.z1_sigma = Lambda(self._get_sigma, name="calculate_sigma_z1")(
            [self.z1_log_sigma_BU, self.z1_log_sigma_TD])

        self.z1_mean = Lambda(self._get_mean, name="calculate_mean_z1")([
            self.z1_mean_TD, self.z1_log_sigma_TD, self.z1_mean_BU,
            self.z1_log_sigma_BU
        ])

        # Samlpe z1!
        self.z1 = Lambda(self._sample,
                         name="sampling_z1")([self.z1_mean, self.z1_sigma])

        reco = decoder1(self.z1)

        model = Model(input_img, reco)
        model.beta = K.variable(self.beta)

        losses = Losses(loss_fn,
                        self.observation_noise,
                        model.beta,
                        z2_mean=self.z2_mean,
                        z2_log_sigma=self.z2_log_sigma,
                        z_mean=self.z1_mean,
                        z_sigma=self.z1_sigma,
                        z_mean_TD=self.z1_mean_TD,
                        z_log_sigma_TD=self.z1_log_sigma_TD)

        model.compile(optimizer=RMSprop(lr=lr, decay=decay),
                      loss=losses.loss,
                      metrics=[losses.KL_divergence])

        # Generative model
        latent_input = Input(shape=(self.latent_dim2, ))
        gen2 = decoder2(latent_input)

        # Using same TD mean sigma generator as before
        gen_mean, gen_log_sigma = mean_log_sigma_model_for_top_down_calc(gen2)

        # Combine mean and sigma for generative model
        gen_sigma = Lambda(self._get_sigma_gen)([gen_log_sigma])

        gen_mean = Lambda(self._get_mean_gen)([gen_mean, gen_log_sigma])

        # Samlping
        gen2 = Lambda(self._sample)([gen_mean, gen_sigma])

        gen_reco = decoder1(gen2)
        generative_model = Model(latent_input, gen_reco)

        # Model for latent inference
        z1 = self.z1
        z1_mean = self.z1_mean
        z1_sigma = self.z1_sigma
        z2 = self.z2
        z_mean_BU = self.z1_mean_BU
        z_log_sigma_BU = self.z1_log_sigma_BU
        z_mean_TD = self.z1_mean_TD
        z_log_sigma_TD = self.z1_log_sigma_TD
        args = z_mean_BU, z_log_sigma_BU, z_mean_TD, z_log_sigma_TD
        latent_model = Model(input_img,
                             outputs=[
                                 reco, z1, z1_mean, z1_sigma, z2, z_mean_BU,
                                 z_log_sigma_BU, z_mean_TD, z_log_sigma_TD
                             ])

        self.latent_dim = self.latent_dim2

        return model, generative_model, latent_model



from __future__ import absolute_import

from .autoencoders import DenseAutoEncoder, ConvolutionalAutoEncoder
from .csnl_util import DataGenerator, DataGeneratorWithLabels
from .train import ModelTrainer
from .variational_autoencoders import DenseVAE, SmallDenseVAE, ConvolutionalVAE, SmallConvolutionalVAE
from .visualize import VAEPlotter, GifCallBack
from .ladder_variational_autoencoders import ConvLadderVAE, DenseLadderVAE, SmallDenseLadderVAE, SmallDenseLadderVAE_BN, SmallDenseLinearLadderVAE, DenseLinearLadderVAE
from .ladder_variational_autoencoders_bn import ConvLadderVAE_BN, DenseLadderVAE_BN, DenseLinearLadderVAE_BN


import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import numpy as np
from keras.utils import to_categorical
from scipy.stats import pearsonr
import os


class VAEPlotter:
    def __init__(self, fitted_model, datagen, label_datagen, grid_size=14):
        self.model = fitted_model.model
        self.generator_model = fitted_model.generator
        self.latent_model = fitted_model.latent_model
        self._latent_dim = fitted_model.latent_dim
        self.datagen = datagen
        self.label_datagen = label_datagen
        self.grid_size = grid_size
        self.train_images, _ = next(self.datagen.flow())
        self.test_images, _ = self.datagen.validation_data()
        self.batch_size = self.train_images.shape[0]
        self.input_shape = self.model.layers[0].output_shape
        self.image_shape = self.train_images.shape[1:]
        if self.image_shape[2] == 1:
            self.image_shape = self.image_shape[:-1]

    def grid_plot(self):
        reco_train = self.model.predict(self.train_images.reshape(
            self.batch_size, *self.input_shape[1:]),
                                        batch_size=self.batch_size)

        reco_test = self.model.predict(
            self.test_images[:self.batch_size].reshape(self.batch_size,
                                                       *self.input_shape[1:]),
            batch_size=self.batch_size)

        reco = [reco_train, reco_test]
        images = [self.train_images, self.test_images]

        fig = plt.figure(figsize=(14, 7))
        outer = gridspec.GridSpec(1, 2, wspace=0.1, hspace=0.1)

        for outer_ind in range(2):
            inner = gridspec.GridSpecFromSubplotSpec(
                4,
                4,
                subplot_spec=outer[outer_ind],
                wspace=0.095,
                hspace=0.095)
            for inner_ind in range(16):
                ax = plt.Subplot(fig, inner[inner_ind])
                if inner_ind % 2 == 0:
                    if inner_ind < 4:
                        ax.set_title('Reconstructed')
                    ax.imshow(reco[outer_ind][inner_ind].reshape(
                        *self.image_shape),
                              interpolation='none',
                              vmin=0,
                              vmax=1)
                else:
                    if inner_ind < 4:
                        ax.set_title(' <- Original')
                    ax.imshow(images[outer_ind][inner_ind - 1].reshape(
                        *self.image_shape),
                              interpolation='none',
                              vmin=0,
                              vmax=1)
                ax.set_xticks([])
                ax.set_yticks([])
                fig.add_subplot(ax)

        fig.suptitle("Train and test reconstructions\n\n")
        plt.savefig(os.getcwd() + "/results/reconstrunction_samples.png",
                    dpi=50)
        plt.show()

    def generate_samples(self, vmax=1):
        latent_inputs = np.random.normal(size=(self.grid_size * self.grid_size,
                                               self._latent_dim))

        self._plot_samples(latent_inputs, vmax)

    def visualize_latent(self, axis=0, sweep_from=-1, sweep_to=1, vmax=1):
        sweep = np.linspace(sweep_from, sweep_to,
                            self.grid_size * self.grid_size)
        latent_inputs = np.random.normal(size=(1, self._latent_dim))
        latent_inputs = np.array(latent_inputs.tolist() *
                                 (self.grid_size * self.grid_size))
        latent_inputs = latent_inputs.reshape(self.grid_size * self.grid_size,
                                              self._latent_dim)
        latent_inputs[:, axis] = sweep

        self._plot_samples(latent_inputs, vmax)

    def _plot_samples(self, latent_inputs, vmax):
        recos = self.generator_model.predict(
            latent_inputs[:self.grid_size * self.grid_size].reshape(
                self.grid_size * self.grid_size, self._latent_dim))
        # Output 196 images
        fig, axes = plt.subplots(self.grid_size,
                                 self.grid_size,
                                 sharex=True,
                                 sharey=True,
                                 figsize=(11, 11))

        if np.prod(recos[0].shape) / (28 * 28) != 1:
            recos = recos.reshape(self.grid_size * self.grid_size, 28, 28,
                                  int(np.prod(recos[0].shape) / (28 * 28)))
        else:
            recos = recos.reshape(self.grid_size * self.grid_size, 28, 28)

        for ind, ax in enumerate(axes.flatten()):
            ax.imshow(recos[ind], interpolation='none', vmin=0, vmax=vmax)
            ax.set_xticks([])
            ax.set_yticks([])

        fig.suptitle("Generated samples on %d - dimensional grid" %
                     self._latent_dim)
        plt.savefig(os.getcwd() + "/results/generated_samples.png", dpi=50)
        plt.show()

    def plot_td_bu_values(self, latent_dim1):
        images, _ = next(self.datagen.flow())
        for IMAGE_INDEX in range(images.shape[0]):
            _images = np.copy(images)
            for ind in range(
                    images.shape[0]):  # predict same element of the batch
                _images[ind] = images[IMAGE_INDEX]

            try:
                recos, z1, z1_mean, z1_sigma, z2, z_mean_BU, z_log_var_BU, z_mean_TD, z_log_var_TD = self.latent_model.predict(
                    _images, batch_size=self.batch_size)
            except ValueError:
                recos, z1, z1_mean, z1_sigma, z2, z_mean_BU, z_log_var_BU, z_mean_TD, z_log_var_TD = self.latent_model.predict(
                    _images.reshape(self.batch_size, 28 * 28),
                    batch_size=self.batch_size)

            self._plot_td_bu_comparisons(z_mean_BU, z_log_var_BU, z_mean_TD,
                                         z_log_var_TD, _images, z1_mean,
                                         z1_sigma, recos, IMAGE_INDEX)

    def _plot_td_bu_comparisons(self, z_mean_BU, z_log_var_BU, z_mean_TD,
                                z_log_var_TD, _images, z1_mean, z1_sigma,
                                recos, IMAGE_INDEX):
        mean_and_vars = [[z_mean_BU, z_log_var_BU], [z_mean_TD, z_log_var_TD]]
        names = ['Bottom up values', 'Top down values']
        for img_ind in range(_images.shape[0]):
            fig, axes = plt.subplots(3,
                                     2,
                                     sharex=False,
                                     sharey=False,
                                     figsize=(7, 9))
            for ind in range(0, 2):
                mean, log_var = mean_and_vars[ind]
                axes[ind, 0].hist(z1_mean[img_ind],
                                  bins=10,
                                  alpha=0.2,
                                  label="z1")
                axes[ind, 1].hist(z1_sigma[img_ind],
                                  bins=10,
                                  alpha=0.2,
                                  label="z1")
                axes[ind, 0].hist(mean[img_ind],
                                  bins=10,
                                  alpha=0.2,
                                  label="other")
                axes[ind, 1].hist(np.exp(log_var[img_ind]),
                                  bins=10,
                                  alpha=0.2,
                                  label="other")
                axes[ind, 0].set_title(names[ind] + " - mean")
                axes[ind, 1].set_title(names[ind] + " - sigma")
                axes[ind, 0].legend(loc='upper right')
                axes[ind, 1].legend(loc='upper right')
            axes[2, 0].imshow(_images[img_ind].reshape(28, 28))
            axes[2, 0].set_title('Original image')
            axes[2, 1].imshow(recos[img_ind].reshape(28, 28))
            axes[2, 1].set_title('Reconstructed image')
            fig.tight_layout()
            plt.savefig(os.getcwd() + "/results/%d_TD_BU_COMPS_%d.png" %
                        (IMAGE_INDEX + 1, img_ind + 1),
                        dpi=50)
            plt.close(fig)

    def plot_label_correlations(self):
        images, labels = next(self.label_datagen.flow())
        try:
            recos, z1, z1_mean, z1_sigma, z2, _, _, _, _ = self.latent_model.predict(
                images, batch_size=self.batch_size)
        except ValueError:
            recos, z1, z1_mean, z1_sigma, z2, _, _, _, _ = self.latent_model.predict(
                images.reshape(self.batch_size, 28 * 28),
                batch_size=self.batch_size)

        if np.prod(images[0].shape) / (28 * 28) != 1:
            images = images.reshape(self.batch_size, 28, 28,
                                    int(np.prod(images[0].shape) / (28 * 28)))
        else:
            images = images.reshape(self.batch_size, 28, 28)

        labels = to_categorical(labels)

        for cat in range(labels[0].shape[0]):
            fig = plt.figure(figsize=(10, 8))
            correlations = []

            for i in range(self._latent_dim):
                correlations.append(pearsonr(z2[:, i], labels[:, cat]))

            correlations = np.array(correlations)

            plt.title("#%d one-hot category correlation with z2" % (cat + 1))
            plt.scatter(range(correlations.shape[0]),
                        correlations[:, 0],
                        c=[
                            'b' if x >= 0.4 else 'r'
                            for x in np.abs(correlations[:, 0])
                        ])

            plt.hlines(y=.4, xmin=0, xmax=self._latent_dim)
            plt.hlines(y=-.4, xmin=0, xmax=self._latent_dim)
            plt.xlim(0, self._latent_dim)
            plt.xticks(np.arange(1, self._latent_dim, 2))
            plt.xlabel('Latent parameters')
            plt.ylabel('Pearson-correlation')
            plt.savefig(os.getcwd() + "/results/cat-%d-to-z2-corr.png" %
                        (cat + 1),
                        dpi=50)
            plt.show()

    def plot_contrast_correlations(self, latent_dim2=None):
        def contrast_flow(_flow):
            def train_generator(_it):
                while True:
                    batch_x, batch_y = next(_it)
                    yield _contrast(batch_x)

            return train_generator(_flow)

        def conv_contrast_flow(_flow):
            def train_generator(_it):
                while True:
                    batch_x, batch_y = next(_it)
                    batch_x, batch_y = _contrast(batch_x)
                    yield batch_x.reshape(self.batch_size, 28, 28,
                                          1), batch_y.reshape(
                                              self.batch_size, 28, 28, 1)

            return train_generator(_flow)

        random_contrasts = []

        def _contrast(images):
            contrasted_images = np.zeros(shape=images.shape)
            for ind in range(images.shape[0]):
                contrast = np.random.rand() * 2.
                random_contrasts.append(contrast)
                contrasted_images[ind] = np.clip(
                    contrast * (images[ind] - 0.5) + 0.5, 0, 1)
            return contrasted_images.reshape(self.batch_size,
                                             784), contrasted_images.reshape(
                                                 self.batch_size, 784)

        try:
            reco, z1, z1_mean, z1_sigma, z2, _, _, _, _ = self.latent_model.predict_generator(
                contrast_flow(self.datagen.flow()), steps=50, verbose=1)
        except ValueError:
            reco, z1, _, _, z2, _, _, _, _ = self.latent_model.predict_generator(
                conv_contrast_flow(self.datagen.flow()), steps=50, verbose=1)

        random_contrasts = np.array(random_contrasts).flatten()[:reco.shape[0]]
        reco = reco.reshape(reco.shape[0], 28, 28)

        self._latent_correlation(z2, random_contrasts, "z2")
        self._stats(z2, self._latent_dim, "z2")
        if latent_dim2 != None:
            self._latent_correlation(z1, random_contrasts, "z1", latent_dim2)
            self._stats(z1, latent_dim2, "z1")

    def _latent_correlation(self,
                            z,
                            random_contrasts,
                            latent_name,
                            latent_dim=None):
        fig = plt.figure(figsize=(10, 8))

        correlations = []

        latent_dim = self._latent_dim if latent_dim == None else latent_dim

        for i in range(latent_dim):
            correlations.append(pearsonr(z[:, i], random_contrasts))

        correlations = np.array(correlations)

        plt.title("Contrast correlation with %s" % latent_name)
        plt.scatter(
            range(correlations.shape[0]),
            correlations[:, 0],
            c=['b' if x >= 0.4 else 'r' for x in np.abs(correlations[:, 0])])
        plt.hlines(y=.4, xmin=0, xmax=latent_dim)
        plt.hlines(y=-.4, xmin=0, xmax=latent_dim)
        plt.xlim(0, latent_dim)
        plt.xticks(np.arange(1, latent_dim, 2))
        plt.xlabel('Latent parameters')
        plt.ylabel('Pearson-correlation')
        plt.savefig(os.getcwd() +
                    "/results/contrast-to-%s-corr.png" % latent_name,
                    dpi=50)
        plt.show()

    def _stats(self, z, latent_dim, name):
        fig = plt.figure(figsize=(20, 10))

        plt.errorbar(x=range(latent_dim),
                     y=np.mean(z, axis=0).flatten(),
                     yerr=np.std(z, axis=0).flatten(),
                     ecolor='r',
                     fmt='o')
        plt.xticks(np.linspace(0, latent_dim, 17))
        plt.title('Mean and standard deviation of %s' % name)
        plt.savefig(os.getcwd() + "/results/mean-and-std-of-%s.png" % name,
                    dpi=50)
        plt.show()



import pickle
from sklearn.utils import shuffle
from sklearn import model_selection
import numpy as np
import os
from sklearn.preprocessing import StandardScaler
from scipy.linalg import svd


class DataLoader:
    def __init__(self,
                 file_path,
                 binarize,
                 zca_whiten=False,
                 contrast_normalize=False):
        self.binarize = binarize
        self.zca_whiten = zca_whiten
        self.contrast_normalize = contrast_normalize
        with open(file_path, 'rb') as f:
            self.data = pickle.load(f)

    def get_data_and_labels(self):
        X_train = self.data['train_images']
        X_test = self.data['test_images']
        y_train = self.data['train_labels']
        y_test = self.data['test_labels']

        X_train = np.clip(X_train, 0, 1.0)
        X_test = np.clip(X_test, 0, 1.0)

        print('Training size : %d \t Test size : %d' %
              (X_train.shape[0], X_test.shape[0]))

        channel = int(np.prod(X_train.shape) / (X_train.shape[0] * 28 * 28))

        # not possible to not split them
        X_train = shuffle(X_train.reshape(X_train.shape[0], 28, 28, channel),
                          random_state=42)
        y_train = shuffle(y_train, random_state=42)
        X_test = shuffle(X_test.reshape(X_test.shape[0], 28, 28, 1),
                         random_state=137)
        y_test = shuffle(y_test, random_state=42)

        if self.binarize:
            X_train[X_train >= .5] = 1.
            X_train[X_train < .5] = 0.
            X_test[X_test >= .5] = 1.
            X_test[X_test < .5] = 0.
        if self.zca_whiten:
            # add +1 and divide by 2 to shift
            # them back to the interval [-1, 1]
            X_train = self._whiten(X_train)
            X_test = self._whiten(X_test)
        if self.contrast_normalize:
            X_train = np.array([
                self._contrast_normalization_train(x, y_train[ind])
                for ind, x in enumerate(X_train)
            ])
            X_test = np.array([
                self._contrast_normalization_test(x, y_test[ind])
                for ind, x in enumerate(X_test)
            ])

        print("Shapes : ", X_train.shape, "\t", X_test.shape)
        print("Label shaped : ", y_train.shape, "\t", y_test.shape)

        mean, std = np.mean(X_train), np.std(X_train)
        print("Train set : ")
        print('Mean: %.3f, Standard Deviation: %.3f' % (mean, std))
        print('Min: %.3f, Max: %.3f' % (np.min(X_train), np.max(X_train)))

        mean, std = np.mean(X_test), np.std(X_test)
        print("Test set : ")
        print('Mean: %.3f, Standard Deviation: %.3f' % (mean, std))
        print('Min: %.3f, Max: %.3f' % (np.min(X_test), np.max(X_test)))

        return (X_train, y_train, X_test, y_test)

    def train_test_split(self):
        try:
            X_train, y_train, X_test, y_test = self.get_data_and_labels()
            return (X_train, X_test)
        except KeyError:
            print('not getting here')
            X_train = self.data['train_images']
            X_test = self.data['test_images']

            X_train = np.clip(X_train, 0, 1.0)
            X_test = np.clip(X_test, 0, 1.0)

            print('Training size : %d \t Test size : %d' %
                  (X_train.shape[0], X_test.shape[0]))

            channel = int(
                np.prod(X_train.shape) / (X_train.shape[0] * 28 * 28))

            # not possible to not split them
            X_train = shuffle(X_train.reshape(X_train.shape[0], 28, 28,
                                              channel),
                              random_state=42)
            X_test = shuffle(X_test.reshape(X_test.shape[0], 28, 28, 1),
                             random_state=137)

            print("Shapes : ", X_train.shape, "\t", X_test.shape)
        except IndexError:
            X = self.data
            X = np.clip(X, 0, 1.0)

            channel = int(np.prod(X.shape) / (X.shape[0] * 28 * 28))

            X_train, X_test = model_selection.train_test_split(
                X.reshape(X.shape[0], 28, 28, channel),
                test_size=0.05,
                random_state=137)
        finally:
            if self.binarize:
                X_train[X_train >= .5] = 1.
                X_train[X_train < .5] = 0.
                X_test[X_test >= .5] = 1.
                X_test[X_test < .5] = 0.
            if self.zca_whiten:
                # add +1 and divide by 2 to shift
                # them back to the interval [-1, 1]
                X_train = self._whiten(X_train)
                X_test = self._whiten(X_test)

            mean, std = np.mean(X_train), np.std(X_train)
            print("Train set : ")
            print('Mean: %.3f, Standard Deviation: %.3f' % (mean, std))
            print('Min: %.3f, Max: %.3f' % (np.min(X_train), np.max(X_train)))

            return (X_train, X_test)

    def _whiten(self, x):
        flat_x = np.reshape(x,
                            (x.shape[0], x.shape[1] * x.shape[2] * x.shape[3]))
        sigma = np.dot(flat_x.T, flat_x) / flat_x.shape[0]
        u, s, _ = svd(sigma)
        s_inv = 1. / np.sqrt(s[np.newaxis] + 1e-12)
        principal_components = (u * s_inv).dot(u.T)
        flatx = np.reshape(x, (-1, np.prod(x.shape[-3:])))
        whitex = np.dot(flatx, principal_components)
        x = np.reshape(whitex, x.shape)
        return x

    def _contrast_normalization_train(self,
                                      x,
                                      label,
                                      s=3.33,
                                      lmbd=10,
                                      epsilon=1e-2):
        means = {
            0: 0.49618801083712977,
            1: 0.5000000000000002,
            2: 0.5000043854028291,
            3: 0.49999037217872805,
            4: 0.4999999999999999,
            5: 0.49999999999999944,
            6: 0.5000000000000006
        }
        stds = {
            0: 0.2795232304659204,
            1: 0.12972179491334127,
            2: 0.1878776468561605,
            3: 0.1455744995271547,
            4: 0.042093943681260626,
            5: 0.08038296207324777,
            6: 0.11335077635372165
        }
        x_label = means[label]
        x_updated = x - x_label
        x_updated /= stds[label]
        x_updated += 0.5  # set new mean
        mini, maxi = -4.753, 7.455  # from train min and max values
        diff = maxi - mini
        x_updated -= mini
        x_updated /= diff
        x_updated = np.clip(x_updated, 0, 1)
        return x_updated

    def _contrast_normalization_test(self,
                                     x,
                                     label,
                                     s=3.33,
                                     lmbd=10,
                                     epsilon=1e-2):
        means = {
            0: 0.49938174332232904,
            1: 0.4994474854248672,
            2: 0.49953722540249085,
            3: 0.4993921865195907,
            4: 0.4994409264556163,
            5: 0.49944513245349165,
            6: 0.4995206918013805
        }
        stds = {
            0: 0.16108819138477978,
            1: 0.15795423837534453,
            2: 0.1530487919614309,
            3: 0.1603416970580439,
            4: 0.15674000765449378,
            5: 0.15705830699788273,
            6: 0.15216802897686568
        }
        x_label = means[label]
        x_updated = x - x_label
        x_updated /= stds[label]
        x_updated += 0.5  # set new mean
        mini, maxi = -2.783, 3.789  # from train mean and max values
        diff = maxi - mini
        x_updated -= mini
        x_updated /= diff
        x_updated *= (0.081913499344692 / 0.15216068167985394)
        x_updated += (0.4302916120576671 - 0.26892201834862384)
        x_updated = np.clip(x_updated, 0, 1)
        return x_updated
